from .abstract import DataSource, ContentSupplier
from .globs import globs_to_pattern
from pydantic.dataclasses import dataclass
from typing import Tuple, Iterable, Generator, Callable, Pattern, Optional, Sequence
from os import walk, listdir
from os.path import isfile, isdir, join, split, relpath
from functools import partial
from fnmatch import translate


def _id(x):
    return x

@dataclass
class InlineDataSource(DataSource):
    collection_of_files: Tuple[Tuple[str, bytes], ...]

    def provide_files(self) -> Iterable[Tuple[str, ContentSupplier]]:
        return tuple((file_name, partial(_id, buffer))
                    for (file_name, buffer) in self.collection_of_files)

def _gen_files(starting_path: str, is_recursive: bool) -> Generator[Tuple[str, str], None, None]:
    if isfile(starting_path):
        if is_recursive:
            raise ValueError('Cannot be recursive over a file')
        yield (split(starting_path)[1], starting_path)

    elif isdir(starting_path):
        if is_recursive:
            source = (join(dir_path, file_name)
                      for (dir_path, _ignore, file_names) in walk(starting_path)
                      for file_name in file_names)
        else:
            source = (
                full_path for full_path in 
                (join(starting_path, file_name) for file_name in listdir(starting_path))
                if isfile(full_path))

        for source_path in source:
            archive_path = relpath(source_path, starting_path)
            yield (archive_path, source_path)

    else:
        raise ValueError(f"Unknown or invalid path {starting_path}")

def _read_file(file_name: str) -> bytes:
    with open(file_name, 'rb') as file_object:
        return file_object.read()

@dataclass
class FileDataSource(DataSource):
    scan_root: str
    is_recursive: bool = True
    whitelist_globs: Sequence[str] = ()
    blacklist_globs: Sequence[str] = ()

    def _create_predicate(self) -> Callable[[str], bool]: 
        whitelist_pattern = globs_to_pattern(self.whitelist_globs)
        blacklist_pattern = globs_to_pattern(self.blacklist_globs)
        
        def _predicate(tested: str):
            is_whitelist_match = ((whitelist_pattern is None) or 
                     (whitelist_pattern.search(tested) is not None))
            is_blacklist_non_match = ((blacklist_pattern is None) or
                     (not blacklist_pattern.search(tested)))
            return is_whitelist_match and is_blacklist_non_match

        return _predicate

    def provide_files(self) -> Iterable[Tuple[str, ContentSupplier]]:
        predicate = self._create_predicate()

        return tuple(
            (archive_path, partial(_read_file, full_path))
            for (archive_path, full_path) in _gen_files(self.scan_root, self.is_recursive)
            if predicate(full_path)
        )


#!/usr/bin/env python
import click
from typing import Tuple, Any, BinaryIO
from .data_sources import InlineDataSource, FileDataSource
from .builder import SharikBuilder

def _process_inline(_ctx: click.Context, _param: Any, values: Tuple[str]) -> Tuple[Tuple[str, bytes], ...]:
    def _gen_element(value):
        first_equals = value.find('=')
        if first_equals <= 0:
            raise click.BadParameter(f'Must be of the form key=value')
        return value[:first_equals], value[first_equals+1:].encode('utf-8')
    return tuple(_gen_element(value) for value in values)

@click.command()
@click.option('-x', 'trace', is_flag=True, required=False, default=False, 
              help="Print each step to standard output")
@click.option('--command', type=str, required=True, help="Shell command to be run after unpacking")
@click.option('--add', type=click.Path(exists=True), multiple=True, help="File to be added")
@click.option('--inline', type=str, multiple=True, help="Inline parameters to be added",
              callback=_process_inline)
@click.option('--clear-glob', type=str, multiple=True, 
              help="Files to be represented without compression/encoding")
@click.option('-o', '--output', type=click.File(mode='wb'), required=True,
              help="File to which to output the result, can be - for stdout")
def cli_main(trace: bool = False,
             command: str = '/bin/false', 
             add: Tuple[str, ...] = (),
             inline: Tuple[Tuple[str, bytes], ...] = (),
             clear_glob: Tuple[str, ...] = (),
             output: BinaryIO = None):
    builder = SharikBuilder(command.encode('utf-8'), trace)
    for element in add:
        builder.add_data_source(FileDataSource(element))
    if len(inline) > 0:
        builder.add_data_source(InlineDataSource(inline))
    for element in clear_glob:
        builder.add_clear_glob(element)
    output.write(builder.build())

if __name__ == '__main__':
    cli_main()



from unittest import TestCase
from sharik.builder import SharikBuilder
from sharik.data_sources import InlineDataSource, FileDataSource

from sys import stderr
from os.path import dirname, basename, join

_OWN_FILE=basename(__file__)
_OWN_DIR=dirname(__file__)
_PARENT_DIR=join(_OWN_DIR, '..')


class TestSanity(TestCase):
    def testSanity(self):
        f = SharikBuilder(b'ls -la')
        f.add_data_source(InlineDataSource((('foo', b'hello world' * 33),)))
        f.add_data_source(InlineDataSource((('bar', b'hello world'),)))
        f.add_data_source(InlineDataSource((('dir/dfs', b'hakuna matata'),)))

        f.add_clear_glob('dir/*')
        stderr.write(f.build().decode('utf-8'))

    def testPrefix(self):
        f = SharikBuilder(b'ls -la')
        f.add_data_source(InlineDataSource((('foo', b'a' * 33),)))
        f.add_data_source(InlineDataSource((('foo', b'b' * 33),)), prefix='test/')
        f.add_clear_glob('foo')

        result = f.build().decode('utf-8')

        self.assertIn('a' * 33, result)
        self.assertIn('b' * 33, result)
        self.assertIn('test/foo', result)

    def testRecursive(self):
        for recursive_status in (False, True):
            f = SharikBuilder(b'ls -la')
            f.add_data_source(FileDataSource(_PARENT_DIR, recursive_status, ('*.py',), ()))
            result = f.build().decode('utf-8')
            self.assertIn('abstract.py', result)
            self.assertIn('data_sources.py', result)
            self.assertEqual(_OWN_FILE in result, recursive_status)

    def testWhitelist(self):
        f = SharikBuilder(b'ls -la')
        anchor = 'data_sources.py'
        f.add_data_source(FileDataSource(_PARENT_DIR, False, (anchor,), ()))
        result = f.build().decode('utf-8')
        self.assertIn(anchor, result)

    def testBlacklist(self):
        f = SharikBuilder(b'ls -la')
        anchor_out = 'data_sources.py'
        anchor_in = 'builder.py'
        f.add_data_source(FileDataSource(_PARENT_DIR, False, ('*.py',), (anchor_out,)))
        result = f.build().decode('utf-8')
        self.assertNotIn(anchor_out, result)            
        self.assertIn(anchor_in, result)

    def testType(self):
        try:
            SharikBuilder('ls -la')
            self.fail("Should not have accepted a parameter that's not bytes")
        except:
            pass


from setuptools import setup, find_packages

setup(
    name="sharik",
    version="0.2.2",
    packages=find_packages(),
    description='A shar(1)-like utility with a programmatic fine-tuned API',
    author='Uri Yanover',
    author_email='throwaway+sharik@yanover.name',
    url='https://github.com/uri-yanover/sharik',
    entry_points={
        'console_scripts': [
            'sharik = sharik.cli:cli_main'
        ]
    },
    install_requires=['click', 'dataclasses', 'pydantic']
)


from abc import ABC, abstractmethod
from pydantic.dataclasses import dataclass
from typing import Iterable, Tuple, Callable

ContentSupplier = Callable[[], bytes]


@dataclass
class DataSource(ABC):
    @abstractmethod
    def provide_files(self) -> Iterable[Tuple[str, ContentSupplier]]:
        pass





from functools import partial
from pydantic.dataclasses import dataclass
from dataclasses import field
from collections import OrderedDict
from os import path
from typing import Callable, Dict, Generator, Tuple, List, NoReturn, Iterable
from base64 import b64encode
from fnmatch import fnmatch
from zlib import compressobj, MAX_WBITS

from .abstract import DataSource, ContentSupplier
from .globs import globs_to_pattern

# Note that '-' is not a valid base64 character
_END_MARKER_STR = '---END-SHARIK---'
_END_MARKER_BYTES = _END_MARKER_STR.encode('utf-8')
_END_MARKER_CONDITION_BYTES = b'\n' + _END_MARKER_BYTES
_ZLIB_COMPRESSION_LEVEL = 9

# See: https://stackoverflow.com/questions/1838699/how-can-i-decompress-a-gzip-stream-with-zlib
def _compress(data: bytes) -> bytes:
    compress_object = compressobj(level=_ZLIB_COMPRESSION_LEVEL, wbits = MAX_WBITS | 16)
    result = compress_object.compress(data)
    return result + compress_object.flush()

@dataclass
class _SharikShellGenerator(object):
    final_command: bytes
    trace: bool
    clear_globs: List[str]
    elements: Tuple[Tuple[str, ContentSupplier], ...]

    def _gen_header(self) -> Generator[bytes, None, None]:
        yield b'#!/bin/sh'

        if self.trace:
           yield b'set +x'
        
        yield f'''decode() {{
     base64 --decode | gunzip - > "${1}"
}}'''.encode('utf-8')

    @staticmethod
    def _gen_per_file(name: str, 
                      is_clear: bool, 
                      contents: ContentSupplier) -> Generator[bytes, None, None]:
        directory_name = path.split(name)[0]
        un_encoded_bytes = contents()

        if directory_name is not None and directory_name != '':
            yield f'mkdir -p {directory_name}'.encode('utf-8')

        if is_clear:
            if _END_MARKER_CONDITION_BYTES in un_encoded_bytes:
                raise ValueError(f"File {name} should not contain the end sequence {_END_MARKER_STR}")
            yield f'cat > {name} <<"{_END_MARKER_STR}"'.encode('utf-8')
            yield un_encoded_bytes
            yield _END_MARKER_BYTES
        else:
            encoded_bytes = b64encode(_compress(un_encoded_bytes))

            yield f'decode {name} <<"{_END_MARKER_STR}"'.encode('utf-8')
            yield encoded_bytes
            yield _END_MARKER_BYTES

    def _gen(self) -> Generator[bytes, None, None]:
        yield from self._gen_header()

        known_files = set()

        pattern = globs_to_pattern(self.clear_globs)
        for (file_name, supplier) in self.elements:
            if file_name in known_files:
                raise ValueError(f"Received the same file twice {file_name}")
            known_files.add(file_name)
            is_clear = pattern is not None and pattern.search(file_name) is not None
            yield from self._gen_per_file(file_name, is_clear, supplier)
        
        # Do not separate those two statements!
        yield self.final_command
        yield b'exit $?'

    def gen_bytes(self) -> bytes:
        return b'\n'.join(self._gen())

@dataclass
class DataSourceWithPrefix(object):
    data_source: DataSource
    prefix: str

    def get_files(self) -> Iterable[Tuple[str, ContentSupplier]]:
        return ((self.prefix + file_name, content_supplier) for 
                (file_name, content_supplier) in self.data_source.provide_files())

@dataclass
class SharikBuilder(object):
    final_command: str
    trace: bool = False
    components: List[DataSourceWithPrefix] = field(default_factory=list)
    clear_globs: List[str] = field(default_factory=list)

    def add_data_source(self, data_source: DataSource, prefix: str='') -> NoReturn:
        self.components.append(DataSourceWithPrefix(data_source, prefix))

    def add_clear_glob(self, clear_glob: str) -> NoReturn:
        self.clear_globs.append(clear_glob)

    def _provide(self, name: str) -> bytes:
        return self.components[name]

    def normalized(self) -> Tuple[Tuple[str, ContentSupplier], ...]:
        result = []
        
        file_name_to_data_source = {}
        for data_source_with_prefix in self.components:
            for (file_name, content_supplier) in data_source_with_prefix.get_files():
                result.append((file_name, content_supplier))
                previous_source = file_name_to_data_source.get(file_name)
                if previous_source is not None:
                    raise ValueError(f"More than one source for the file name {file_name}. " \
                                     f"Previous source: {previous_source}, current one {data_source_with_prefix}")
                file_name_to_data_source[file_name] = data_source_with_prefix

        result.sort(key=lambda pair: pair[0])
        return tuple(result)

    def build(self) -> bytes:
        return _SharikShellGenerator(
            self.final_command,
            self.trace,
            self.clear_globs, 
            self.normalized()).gen_bytes()
